--!parent
--!encoding=utf-8

--****************************************************************************
-- GestureRecognizer
--
-- @author Valentin Schmidt
-- @version 0.1
--
-- Supported gestures:
-- * tap (tap, doubleTap, longTap)
-- * pan (panStart, pan, panStop) (=drag)
-- * pinch (pinchStart, pinch, pinchStop) (=zoom)
-- * rotate (rotateStart, rotate, rotateStop)
-- * swipe (#left, #right, #top, #bottom)
--****************************************************************************

global $

property SWIPE_DIST_MIN
property SWIPE_SPEED_MIN

property TAP_TIME_MAX
property DOUBLETAP_TIME_MAX
property LONGTAP_TIME

property PINCH_DIST_MIN
property PINCH_GUTTER_DIST_MIN

property ROTATION_DETECT_ANGLE_MIN
property ROTATION_DETECT_ANGLE_MAX

property _listeners
property _gesture

property _touchPoints
property _timeLastTouch

property _longTapCheckTimer
property _longTapCheckID

property _angleStart
property _distStart

----------------------------------------
--
----------------------------------------
on new (me)

	$.import("multitouch", me)

	SWIPE_DIST_MIN = 200
	SWIPE_SPEED_MIN = 2 -- pixel per ms
	TAP_TIME_MAX = 300
	DOUBLETAP_TIME_MAX = 200
	LONGTAP_TIME = 1000
	PINCH_DIST_MIN = 10
	PINCH_GUTTER_DIST_MIN = 50
	-- used when distinguishing rotate from pinch
	ROTATION_DETECT_ANGLE_MIN = 45 * PI/180
	ROTATION_DETECT_ANGLE_MAX = 135 * PI/180

	me._listeners = [:]
	me._listeners[#tap] = []
	me._listeners[#pan] = []
	me._listeners[#pinch] = []
	me._listeners[#rotate] = []
	me._listeners[#swipe] = []

	me._touchPoints = [:]
	me._timeLastTouch = 0
	me._gesture = #none

	-- timout used for long-tap checking
	me._longTapCheckTimer = timeout().new("__LONGTAP__", 0, #_done, #_checkLongTap, me)

	$.multitouch.listen_start(TRUE)

	return me
end

----------------------------------------
--
----------------------------------------
on reset (me)
	me._touchPoints = [:]
end

----------------------------------------
--
----------------------------------------
on registerEvent (me, event, listener)
	me._listeners[event].add(listener)
end

----------------------------------------
--
----------------------------------------
on unregisterEvent (me, event, listener)
	me._listeners[event].deleteOne(listener)
end

----------------------------------------
--
----------------------------------------
on updateLocZ (me)
	repeat with event in [#tap, #pan, #pinch, #swipe]
		me._sortByLocZ(event)
	end repeat
end

----------------------------------------
--
----------------------------------------
on touchStart (me, id, p)
	id = string(id)
	if me._touchPoints.count>1 then me._touchPoints.deleteAt(1)

	-- time: used for tap detection
	-- point_last: used for pan
	me._touchPoints[id] = [#point_start:p, #point_last:p, #point_now:p, #time:_system.milliseconds]

	if me._touchPoints.count=1 then -- single touch point: pan or swipe?
		me._gesture = #pan
		me._emit(#panStart, #pan, p)

		-- start checking for long tap
		me._longTapCheckID = id
		me._longTapCheckTimer.period = LONGTAP_TIME

	else -- 2 touch points: pinch (zoom) or rotate?
		me._longTapCheckTimer.period = 0
		me._gesture = #check
	end if
end

----------------------------------------
--
----------------------------------------
on touchMove (me, id, p)
	id = string(id)
	if voidP(me._touchPoints[id]) then return

	me._touchPoints[id][#point_now] = p

	if me._gesture = #check then me._checkGestureType()

	case me._gesture of
		#pan: me._handlePan()
		#pinch: me._handlePinch()
		#rotate: me._handleRotate()
	end case
end

----------------------------------------
--
----------------------------------------
on touchEnd (me, id, p)
	id = string(id)
	if voidP(me._touchPoints[id]) then return

	if me._touchPoints.count=1 then
		ms = _system.milliseconds

		if ms-me._timeLastTouch<=DOUBLETAP_TIME_MAX then
			me._emit(#doubleTap, #tap, p)

		else if ms-me._touchPoints[1][#time]<=TAP_TIME_MAX then
			me._emit(#tap, #tap, p)
			me._timeLastTouch = ms

		else if me._gesture = #pan then
			me._emit(#panStop, #pan)
		end if

	else if me._gesture = #pinch then
		me._emit(#pinchStop, #pinch)

	else if me._gesture = #rotate then
		me._emit(#rotateStop, #rotate)

	end if

	me._gesture = #none
	me._touchPoints.deleteProp(id)

	me._longTapCheckTimer.period = 0
end

----------------------------------------
--
----------------------------------------
on _handlePan (me)

	-- check if swipe
	if me._listeners[#swipe].count then
		dt = _system.milliseconds-me._touchPoints[1].time
		if dt>0 then

			-- check if horizontal swipe
			x = me._touchPoints[1].point_now.locH
			dx = float(abs(x-me._touchPoints[1].point_start.locH))
			if dx>=SWIPE_DIST_MIN and dx/dt>=SWIPE_SPEED_MIN then
				if x>me._touchPoints[1].point_start.locH then
					me._emit(#swipe, #swipe, #right)
				else
					me._emit(#swipe, #swipe, #left)
				end if
				me._touchPoints = [:]
				return
			end if

			-- check if vertical swipe
			y = me._touchPoints[1].point_now.locV
			dy = float(abs(y-me._touchPoints[1].point_start.locV))
			if dy>=SWIPE_DIST_MIN and dy/dt>=SWIPE_SPEED_MIN then
				if y>me._touchPoints[1].point_start.locV then
					me._emit(#swipe, #swipe, #bottom)
				else
					me._emit(#swipe, #swipe, #top)
				end if
				me._touchPoints = [:]
				return
			end if
		end if
	end if

	-- else it's pan
	me._emit(#pan, #pan, me._touchPoints[1].point_last, me._touchPoints[1].point_now)
	me._touchPoints[1].point_last = me._touchPoints[1].point_now
end

----------------------------------------
--
----------------------------------------
on _handlePinch (me)
	if me._listeners[#pinch].count then
		scale_x = 1.
		if me._touchPoints[1][#horizontal] then
			dist_x_start = me._touchPoints[1].point_start.locH - me._touchPoints[2].point_start.locH
			if abs(dist_x_start)>=PINCH_DIST_MIN then
				dist_x_now = me._touchPoints[1].point_now.locH - me._touchPoints[2].point_now.locH
				scale_x = float(dist_x_now)/dist_x_start
			end if
		end if

		scale_y = 1.
		if me._touchPoints[1][#vertical] then
			dist_y_start = me._touchPoints[1].point_start.locV - me._touchPoints[2].point_start.locV
			if abs(dist_y_start)>=PINCH_DIST_MIN then
				dist_y_now = me._touchPoints[1].point_now.locV - me._touchPoints[2].point_now.locV
				scale_y = float(dist_y_now)/dist_y_start
			end if
		end if

		me._emit(#pinch, #pinch, scale_x, scale_y)
	end if
end

----------------------------------------
--
----------------------------------------
on _handleRotate (me)
	if me._listeners[#rotate].count then
		--angle_now = me._angle(me._touchPoints[2].point_now, me._touchPoints[1].point_start)
		angle_now = me._angle(me._touchPoints[2].point_now, me._touchPoints[1].point_now)
		me._emit(#rotate, #rotate, me._angleStart-angle_now)
	end if
end

----------------------------------------
-- Arcossine
-- @param {float} x
-- @return {float}
----------------------------------------
on _acos (me, x)
	res = atan(sqrt(x*x/(1-x*x)))
	if x<0 then res = -res
	return PI/2 - res
end

----------------------------------------
--
----------------------------------------
on _dist (me, p1, p2)
	--return vector(p1[1],p1[2],0).distanceTo(vector(p2[1],p2[2],0))
	
	-- this is faster!
	dx = p1.locH - p2.locH
	dy = p1.locV - p2.locV
	return sqrt(0. + dx*dx + dy*dy)
end

----------------------------------------
--
----------------------------------------
on _angle (me, p1, p2)
	dx = p1.locH - p2.locH
	dy = p1.locV - p2.locV
	a = atan(dx, dy) * 180/PI
	--a = atan(dy, dx) * 180/PI
	if a<0 then return a+360
	return a
end

----------------------------------------
-- checks if pinch or rotate
----------------------------------------
on _checkGestureType (me)

	b = me._dist(me._touchPoints[2].point_start, me._touchPoints[2].point_now)

	if b=0 then return
	c = me._dist(me._touchPoints[1].point_start, me._touchPoints[2].point_start)

	if c=0 then return
	a = me._dist(me._touchPoints[1].point_start, me._touchPoints[2].point_now)

	alpha = me._acos((b*b + c*c - a*a)/(2*b*c)) --*180/PI

	if alpha>=ROTATION_DETECT_ANGLE_MIN and alpha<=ROTATION_DETECT_ANGLE_MAX then
		--put "==> ROTATE"
		me._gesture = #rotate
		me._angleStart = me._angle(me._touchPoints[2].point_start, me._touchPoints[1].point_start)
		me._emit(#rotateStart, #rotate, me._touchPoints[1].point_start)
	else
		--put "==> PINCH"
		me._gesture = #pinch
		if me._listeners[#pinch].count then
			if abs(me._touchPoints[1].point_start.locH - me._touchPoints[2].point_start.locH)>=PINCH_GUTTER_DIST_MIN then
				me._touchPoints[1][#horizontal] = TRUE
			end if
			if abs(me._touchPoints[1].point_start.locV - me._touchPoints[2].point_start.locV)>=PINCH_GUTTER_DIST_MIN then
				me._touchPoints[1][#vertical] = TRUE
			end if
			me._emit(#pinchStart, #pinch, me._touchPoints[1].point_start, me._touchPoints[2].point_start)
		end if
	end if
end

----------------------------------------
--
----------------------------------------
on _checkLongTap (me, arg)
	tp = me._touchPoints[me._longTapCheckID]
	if not voidP(tp) then
		if tp[#point_start]=tp[#point_now] then
			me._emit(#longTap, #tap, tp[#point_start])
			me._gesture = #none
		end if
	end if
	me._longTapCheckTimer.period = 0
end

----------------------------------------
--
----------------------------------------
on _sortByLocZ (me, event)
	p = [:]
	p.sort()
	cnt = me._listeners[event].count
	repeat with i = cnt down to 1
		p.addProp(me._listeners[event][i].locZ, me._listeners[event][i])
		me._listeners[event].deleteAt(i)
	end repeat
	repeat with i = cnt down to 1
		me._listeners[event].add(p[i])
	end repeat
end

----------------------------------------
--
----------------------------------------
on _emit (me, event, eventKey, p1, p2)
	repeat with obj in me._listeners[eventKey]
		if call(event, [obj], p1, p2)=TRUE then exit repeat
	end repeat
end
